$ifndef _SD_INTERFACE_
$define _SD_INTERFACE_
'
' SD memory card interface routines
'
' Written by Les Johnson. September 2011
'
Reminders = Off
'
' SD Card Declarations
' Default settings for the Amicus Touch Screen Colour LCD Shield
'
$ifndef SD_SDO
    $define SD_SDO PORTC.5                  ' SPI DO to SD card DI pin
$endif
$ifndef SD_SDI
    $define SD_SDI PORTC.4                  ' SPI DI to SD card DO pin
$endif
$ifndef SD_CLK
    $define SD_CLK PORTC.3                  ' SPI CLK to SD card CLK pin
$endif
$ifndef SD_CS
    $define SD_CS PORTA.4                   ' SPI CS to SD card CS pin (active low)
$endif

$ifndef SD_Power
    $define SD_Power PORTA.5                ' Enable/Disable the SD card's VDD (active low)
$endif

' $define SD_Detect PORTA.2                 ' Detect if the SD card is inserted

'
' SD card commands
'
    Symbol cGO_IDLE_STATE            = 0 | $40
    Symbol cSEND_OP_COND             = 1 | $40
    Symbol cSEND_CSD                 = 9 | $40
    Symbol cSTOP_TRANSMISSION        = 12 | $40
    Symbol cSEND_STATUS              = 13 | $40
    Symbol cSET_BLOCK_LEN            = 16 | $40
    Symbol cREAD_SINGLE_BLOCK        = 17 | $40
    Symbol cREAD_MULTIPLE_BLOCKS     = 18 | $40
    Symbol cWRITE_SINGLE_BLOCK       = 24 | $40
    Symbol cWRITE_MULTIPLE_BLOCKS    = 25 | $40
    Symbol cERASE_BLOCK_START_ADDR   = 32 | $40
    Symbol cERASE_BLOCK_END_ADDR     = 33 | $40
    Symbol cERASE_SELECTED_BLOCKS    = 38 | $40
    Symbol cCRC_ON_OFF               = 59 | $40

    Symbol cDATA_START_TOKEN = $FE  ' An SD card start token
    Symbol cDATA_ACCEPTED = $05     ' An SD card data accepted token
    Symbol cNODATA = 0              ' indicates that the SD card does Not expect To transmit Or receive more data
    Symbol cSD_FLOATING_BUS = $FF   ' A floating SPI bus condition
    Symbol cSD_BAD_RESPONSE = cSD_FLOATING_BUS  ' A bad SD card response byte

$ifndef True
    $define True 1
$endif
$ifndef False
    $define False 0
$endif
'
' Declare variables used for SD interfacing
' Note that some are made system variable so they will reside in Access RAM
'
    Dim SD_Flags As Byte System
    Dim tSD_Response As SD_Flags.0

    Dim SD_bResponseRequired As Byte
    Dim PP7 As Byte System
    Dim PP7H As Byte System
    Dim PP7HH As Byte System
    Dim PP7HHH As Byte System
    Dim SD_dAddress As PP7.Dword
    Dim SD_bCommandResponse As Byte
    Dim SD_bCommandRetry As Byte

    Dim SD_dCurrentSector As Dword System   ' The sector we're currently reading/writing

    Dim SD_wInitRetry As Word System
    Dim SD_bInitRetry As SD_wInitRetry.Byte0
    Dim SD_wSectorLoop As SD_wInitRetry

    Symbol tSD_ZeroFlag = STATUS.2
    Symbol tSD_OverFlowFlag = STATUS.0

    Symbol SPI_OSC16 = %00000001            ' SPI peripheral Master mode, clock = Fosc/16
    Symbol SPI_OSC64 = %00000010            ' SPI peripheral Master mode, clock = Fosc/64

$define tWithIncrement 1                    ' SPI read with counter increment
$define tWithoutIncrement 0                 ' SPI read without counter increment

$define cInserted False                     ' Card insertion result

$define SD_PowerOn() Clear SD_Power         ' Enable power to the SD card (if the circuit allows)
$define SD_PowerOff() High SD_Power         ' Disable power to the SD card (if the circuit allows)

'---------------------------------------------------------------------
    GoTo _SD_Main                           ' Jump over the routines
'-----------------------------------------------------------------------------------
' Set up the MSSP peripheral for master operation
' Parameters    : MSSP peripheral setup values
' Return Value  : None
' Notes         : Set up the MSSP module for master SPI operation
'
' Syntax: SD_SPI_Open(SPI_OSC16 or SPI_OSC64)
'
$define SD_SPI_Open(pSyncMode) _SD_SPI_Open pSyncMode

_SD_SPI_Open Macro- pSyncMode

$ifndef SD_Software_SPI             ' Are we using a software SPI interface?
    T2CON = 0
    SSPSTAT = %10000000             ' No. So input data sampled at end of data output time
                                    ' Output data changes on clock transition from idle to active
    #if(pSyncMode == SPI_OSC64)     ' 1MHz operation?
    SSPCON1 = %00100010             ' Yes. So. Enable the synchronous serial port
                                    ' Clock idle state low
                                    ' SPI Master mode, clock = FOSC/64 (1MHz operation)
    #else                           ' Otherwise...
    SSPCON1 = %00100001             ' Enable the synchronous serial port
                                    ' Clock idle state low
                                    ' SPI Master mode, clock = FOSC/16 (4MHz operation)
    #endif
$endif
Endm

'-----------------------------------------------------------------------------------
' Write a single Byte (8-bits) to the SPI bus more slowly.
' Input     : WREG holds the byte to transmit to the SPI interface
' Output    : None
' Notes     : Used for initialising the SD card
'
$ifdef SD_Software_SPI             ' Are we using a software SPI interface?

$define SD_SPI_WriteSlow(pDataOut) '
    WREG = pDataOut                '
    GoSub __SD_SPI_WriteSlow

__SD_SPI_WriteSlow:
    PRODL = WREG                    ' Transfer WREG into PRODL
    PRODH = 8                       ' Single byte SPI loop
_SPI_WriteSlowLoop:
        SD_SDO = PRODL.7            ' Put current outgoing bit on SD_SDO
        PRODL = PRODL << 1          ' Shift next bit into MSB
        Set SD_CLK                  ' Set SD_CLK high
        DelayCS 100                 ' Slow things down
        Clear SD_CLK                ' Pull SD_CLK low
        DelayCS 100                 ' Slow things down
     Djnz PRODH,_SPI_WriteSlowLoop  ' Do 8-bits
     Return
$endif
'-----------------------------------------------------------------------------------
' Write a single Byte (8-bits) to the SPI bus.
' Input     : WREG holds the byte to transmit to the SPI interface
' Output    : Input data can be picked up by reading WREG straight after the SD_SPI_Write macro
' Notes     : None
'
$define SD_SPI_Write(pDataOut) '
    WREG = pDataOut            '
    GoSub __SD_SPI_Write

__SD_SPI_Write:

$ifdef SD_Software_SPI              ' Are we using a software SPI interface?
    Movwf PRODL                     ' Yes. So transfer WREG into PRODL
    PRODH = 8                       ' Single byte SPI loop
_SPI_WriteLoop:
        SD_SDO = PRODL.7            ' Put current outgoing bit on SD_SDO
        PRODL = PRODL << 1          ' Shift next bit into MSB
        Set SD_CLK                  ' Set SD_CLK high
        Nop
        Clear SD_CLK                ' Pull SD_CLK low
     Djnz PRODH,_SPI_WriteLoop      ' Do 8-bits
     Return
$else                               ' Otherwise. Use the SPI peripheral
    SSPSTATBits_BF = 0              ' Empty the SSPBUF buffer flags
    PIR1bits_SSPIF = 0              ' Clear interrupt flag
    SSPBUF = WREG                   ' Place the data to send into SSPBUF
    Btfss PIR1bits_SSPIF            ' \
    Bra $ - 2                       ' / Wait until cycle complete
    WREG = SSPBUF                   ' Read SSPBUF into WREG
    Return
$endif
'-----------------------------------------------------------------------------------
' Read a single Byte from the SPI bus.
' Input     : WREG contains 1 if a byte counter is to be incremented with the read
' Output    : WREG holds any data read from the SPI interface
' Notes     : None
'
$define SD_SPI_Read(pWithIncrement) _SD_SPI_Read pWithIncrement

_SD_SPI_Read Macro- pWithIncrement \Byte
    Num_Wreg pWithIncrement
    GoSub __SD_SPI_Read
    Return_Wreg
Endm

__SD_SPI_Read:
    If WREG > 0 Then                ' Do we need to increment the byte counter with a read?
        Inc Sector_wByteCounter     ' Yes. So increment the byte counter
    EndIf
$ifdef SD_Software_SPI              ' Are we using a software SPI interface?
    PRODH = 8                       ' Yes. So create a single byte loop
_SPI_ReadLoop:
        PRODL = PRODL << 1          ' Shift next bit into MSB
        Set SD_CLK                  ' Set SD_CLK high
        PRODL = PRODL | SD_SDI      ' Capture current bit on SD_SDI
        Clear SD_CLK                ' Pull SD_CLK low
    Djnz PRODH,_SPI_ReadLoop        ' Do 8-bits
    WREG = PRODL                    ' Transfer PRODL into WREG
    Return
$else                               ' Otherwise. Use the SPI peripheral
    WREG = SSPBUF                   ' Empty the SSPBUF buffer and clear flags
    PIR1bits_SSPIF = 0              ' Clear interrupt flag
    SSPBUF = $FF                    ' Initiate bus cycle
    Btfss PIR1bits_SSPIF            ' \
    Bra $ - 2                       ' / Wait until cycle complete
    WREG = SSPBUF
    Return
$endif
'---------------------------------------------------------------------
' Detect if the card is inserted
' Input     : None
' Output    : SD_Detect is false if the card is inserted
' Notes     : Presumes that the SD Detect pin is already an input
'
$define SD_Present() SD_Detect
'---------------------------------------------------------------------
' Enable the SPI interface for the card
' Input     : None
' Output    : None
' Notes     : Presumes that the CS pin is already an output
'
$define SD_Enable() Clear SD_CS
'---------------------------------------------------------------------
' Disable the SPI interface for the card
' Input     : None
' Output    : None
' Notes     : Presumes that the CS pin is already an output
'
$define SD_Disable() Set SD_CS
'---------------------------------------------------------------------
' Send a command to SD card slowly for card initialisation
' Input     : SD_bResponseRequired holds the required response to wait for
'           : SD_dAddress holds the 32-bit data to send to the card
'           : WREG holds the command to send to the card
' Output    : SD_bCommandResponse holds the response from the SD card
' Notes     : Does not manipulate the SD card's CS line
'
$ifdef SD_Software_SPI                          ' Are we using a software SPI interface?

$define SD_SendCommandSlowly(pCmd, pAddress, pResponseReq) '
    SD_dAddress = pAddress                           '
    SD_bResponseRequired = pResponseReq              '
    WREG = pCmd                                      '
    _SD_SendCommandSlowly

_SD_SendCommandSlowly Macro-
    GoSub __SD_SendCommandSlowly
    Endm

__SD_SendCommandSlowly:
    SD_SPI_WriteSlow(WREG)                      ' Send the command to the SD card, first two bits always '01'
    SD_SPI_WriteSlow(SD_dAddress.Byte3)         ' \
    SD_SPI_WriteSlow(SD_dAddress.Byte2)         ' |
    SD_SPI_WriteSlow(SD_dAddress.Byte1)         ' | Send the data bytes to the SD card
    SD_SPI_WriteSlow(SD_dAddress.Byte0)         ' /
    SD_SPI_WriteSlow($95)                       ' Send the CRC to the SD card, not used
'
' Wait for a response
'
    SD_bCommandRetry = $FF                      ' 255 attempts at the command
    Repeat
        SD_bCommandResponse = SD_SPI_Read(tWithoutIncrement)
        If SD_bCommandResponse = SD_bResponseRequired Then
            Break                               ' Exit the loop if the correct response is received
        EndIf
        Dec SD_bCommandRetry
    Until SD_bCommandRetry = 0
    SD_SPI_WriteSlow($FF)                       ' Send 8 clocks for housekeeping
    Return
$endif
'---------------------------------------------------------------------
' Send a command to SD card
' Input     : SD_bResponseRequired holds the required response to wait for
'           : SD_dAddress holds the 32-bit data to send to the card
'           : WREG holds the command to send to the card
' Output    : SD_bCommandResponse holds the response from the SD card
' Notes     : Does not manipulate the SD card's CS line
'
$define SD_SendCommand(pCmd, pAddress, pResponseReq) '
    SD_dAddress = pAddress                           '
    SD_bResponseRequired = pResponseReq              '
    WREG = pCmd                                      '
    _SD_SendCommand

_SD_SendCommand Macro-
    GoSub __SD_SendCommand
    Endm

#ifMacro- _SD_SendCommand
__SD_SendCommand:
    SD_SPI_Write(WREG)                          ' Send the command to the SD card, first two bits always '01'
    SD_SPI_Write(SD_dAddress.Byte3)             ' \
    SD_SPI_Write(SD_dAddress.Byte2)             ' |
    SD_SPI_Write(SD_dAddress.Byte1)             ' | Send the data bytes to the SD card
    SD_SPI_Write(SD_dAddress.Byte0)             ' /
    SD_SPI_Write($95)                           ' Send the CRC to the SD card, not used
'
' Wait for a response
'
    SD_bCommandRetry = $FF                      ' 255 attempts at the command
    Repeat
        SD_bCommandResponse = SD_SPI_Read(tWithoutIncrement)
        If SD_bCommandResponse = SD_bResponseRequired Then
            Break                               ' Exit the loop if the correct response is received
        EndIf
        Dec SD_bCommandRetry
    Until SD_bCommandRetry = 0
    SD_SPI_Write($FF)                           ' Send 8 clocks for housekeeping
    Return
#endIfMacro-
'---------------------------------------------------------------------
' Initialise the SD card into SPI mode
' Input     : None
' Output    : tSD_Response is false if an error occurs
' Notes     :
'
$define SD_Init() GoSub __SD_Init

__SD_Init:
$ifdef SD_Detect                                ' Are we using the SD detect pin?
    Input SD_Detect                             ' Yes. So make the card detect line an input
$endif
    Low SD_CLK                                  ' Make the clock pin an output low
    Input SD_SDI                                ' Make the SDI pin an input
    Output SD_SDO                               ' Make the SDO pin an output

$ifndef SD_Software_SPI                         ' Are we using a software SPI interface?
    SD_SPI_Open(SPI_OSC64)                      ' No. So open the SPI interface at 1MHz (using a 64MHz crystal)
$endif

$ifdef SD_Power_Cycle                           ' DO we need to power cycle the SD card's VCC line?
    SD_PowerOff()                               ' Yes. So disable the SD card's VDD
    DelayMS 100                                 ' Wait a while
    SD_PowerOn()                                ' Enable the SD card's VDD
    DelayMS 100                                 ' Wait for the SD card to stabilise
$endif

    tSD_Response = False                        ' Default to a bad return
    High SD_CS                                  ' Disable the SPI interface and also set the CS pin to output
    SD_bInitRetry = 0                           ' \
    Repeat                                      ' |
        SD_SPI_Write($FF)                       ' | Send 80 clock cycles
        Inc SD_bInitRetry                       ' |
    Until SD_bInitRetry >= 10                   ' /

    SD_Enable()                                 ' Enable the SPI interface to the SD card
    SD_bInitRetry = $FF
    Repeat
        $ifdef SD_Software_SPI                  ' Are we using a software SPI interface?
            SD_SendCommandSlowly(cGO_IDLE_STATE, 0, 1) ' Send 'reset & go idle' command, response should be 01
        $else
            SD_SendCommand(cGO_IDLE_STATE, 0, 1) ' Send 'reset & go idle' command, response should be 01
        $endif
        Dec SD_bInitRetry                       ' Decrement the retries counter
        If tSD_ZeroFlag = True Then
            SD_Disable()                        ' Disable the SPI interface to the SD card
            Return                              ' Exit the subroutine if the "command retries" time out
        EndIf
    Until SD_bCommandResponse = 1               ' Loop until the correct response from the card is seen

    SD_bInitRetry = $FF
    Repeat
        $ifdef SD_Software_SPI                  ' Are we using a software SPI interface?
            SD_SendCommandSlowly(cSEND_OP_COND, 0, 0) ' Activate card's initialisation process
        $else
            SD_SendCommand(cSEND_OP_COND, 0, 0) ' Activate card's initialisation process
        $endif
        Dec SD_bInitRetry                       ' Decrement the retries counter
        If tSD_ZeroFlag = True Then
            SD_Disable()                        ' Disable the SPI interface to the SD card
            Return                              ' Exit the subroutine if the "command retries" time out
        EndIf
    Until SD_bCommandResponse = 0               ' Loop until the correct response from the card is seen

    $ifdef SD_Software_SPI                      ' Are we using a software SPI interface?
        SD_SPI_WriteSlow($FF)                   ' Allow the SD card to complete its housekeeping
    $else
        SD_SPI_Write($FF)                       ' Allow the SD card to complete its housekeeping
    $endif
    SD_Disable()                                ' Disable the SPI interface to the SD card
$ifndef SD_Software_SPI                         ' Are we using a software SPI interface?
    SD_SPI_Open(SPI_OSC16)                      ' No. So speed up the SPI peripheral's interface (4MHz)
$endif
    tSD_Response = True                         ' Signal a good response
    Return

'---------------------------------------------------------------------
' Point to a particular sector within the SD card
' Input     : SD_dCurrentSector holds the required sector
' Output    : tSD_Response is false if an error occurs
' Notes     : A sector read operation must follow this subroutine
'           : Can also read the remaining bytes from within the sector
'
$define SD_PointToSector(pSector) '
    SD_dCurrentSector = pSector   '
    GoSub __SD_ReadSector
'
' Use this routine for pointing to the sector without first skipping bytes with it
'
$define SD_PointToSectorWithBypass(pSector) '
    SD_dCurrentSector = pSector             '
    GoSub __SD_PointToSectorWithBypass

__SD_ReadSector:
'
' Read and discard the remaining bytes from the sector before pointing to a new one
'
    If Sector_wByteCounter != 512 Then                  ' Is the byte counter 512?
        Sector_wFinishLoop = 512 - Sector_wByteCounter  ' No. So calculate the remaining bytes in the sector
        While Sector_wFinishLoop > 0                    ' Form a loop to read the remaining bytes
            SD_SPI_Read(tWithoutIncrement)              ' Read a byte from the sector
            Dec Sector_wFinishLoop                      ' Decrement the loop counter
        Wend                                            ' Close the loop
    EndIf
__SD_PointToSectorWithBypass:
    SD_SPI_Read(tWithoutIncrement)              ' \ Read the CRC and discard it
    SD_SPI_Read(tWithoutIncrement)              ' /
    SD_SPI_Write($FF)                           ' Allow the SD card to complete its housekeeping
    SD_Disable()                                ' Disable the SD card's SPI interface

    Sector_wByteCounter = 0
    tSD_Response = False                        ' Default to a bad return response
    SD_bInitRetry = $FF                         ' Amount of retries before timing out
    While 1 = 1
        SD_Enable()                             ' Enable the SPI interface to the SD card
        SD_SendCommand(cREAD_SINGLE_BLOCK, SD_dCurrentSector << 9, 0) ' Convert sector number into a byte address by multiplying by 512
        If SD_bCommandResponse = $00 Then
            Break
        Else
            SD_Disable()
        EndIf
        Dec SD_bInitRetry                       ' Decrement the retries counter
        If tSD_ZeroFlag = True Then
            SD_Disable()                        ' Disable the SPI interface to the SD card
            Return
        EndIf
        DelayCS 30                              ' A small delay within the loop
    Wend                                        ' Loop until the correct response from the card is seen
'
' Wait for the start token of data block
'
    SD_bInitRetry = $FF                         ' Amount of retries before timing out
    While 1 = 1
        SD_bCommandResponse = SD_SPI_Read(tWithoutIncrement)
        If SD_bCommandResponse = $FE Then Break
        Dec SD_bInitRetry                       ' Decrement the retries counter
        If tSD_ZeroFlag = True Then
            SD_Disable()                        ' Disable the SPI interface to the SD card
            Return
        EndIf
        DelayCS 30                              ' A small delay within the loop
    Wend                                        ' Loop until the correct response from the card is seen
    tSD_Response = True                         ' Signal a good response
_SD_ReadSectorExit:
    Return

'---------------------------------------------------------------------
_SD_Main:

$endif



